@@ -435,7 +435,7 @@ call_apt_worker (int cmd, int state, char *data, int len,
     }
   else if (pending[cmd].done_callback)
     {
-      fprintf (stderr, "apt-worker command %d already pending\n", cmd);
+      add_log ("apt-worker command %d already pending\n", cmd);
       done_callback (cmd, NULL, done_data);
     }
   else
@@ -454,7 +454,12 @@ call_apt_worker (int cmd, int state, char *data, int len,
 	}
       else
 	{
-	  pending[cmd].data = (char *)g_malloc (len);
+	  /* We need to make sure that pending[cmd].data is not NULL,
+	     since that is the flag that tells
+	     send_pending_apt_worker_cmds that this command is indeed
+	     pending.
+	  */
+	  pending[cmd].data = (char *)g_malloc (len == 0? 1 : len);
 	  pending[cmd].len = len;
 	  memcpy (pending[cmd].data, data, len);
 	}
@@ -669,6 +674,7 @@ void
 apt_worker_get_package_details (const char *package,
 				const char *version,
 				int summary_kind,
+				int state,
 				apt_worker_callback *callback,
 				void *data)
 {
@@ -676,7 +682,7 @@ apt_worker_get_package_details (const char *package,
   request.encode_string (package);
   request.encode_string (version);
   request.encode_int (summary_kind);
-  call_apt_worker (APTCMD_GET_PACKAGE_DETAILS, APTSTATE_DEFAULT,
+  call_apt_worker (APTCMD_GET_PACKAGE_DETAILS, state,
 		   request.get_buf (), request.get_len (),
 		   callback, data);
 }
